package telin;

import gate.AnnotationSet;
import gate.Factory;
import gate.FeatureMap;
import gate.Resource;
import gate.creole.ANNIEConstants;
import gate.creole.AbstractLanguageAnalyser;
import gate.creole.ExecutionException;
import gate.creole.ExecutionInterruptedException;
import gate.creole.ResourceInstantiationException;
import gate.creole.ontology.OResource;
import gate.creole.ontology.OClass;
import gate.creole.ontology.OInstance;
import gate.creole.ontology.AnnotationProperty;
import gate.creole.ontology.Literal;
import gate.creole.ontology.OConstants;
import gate.creole.ontology.Ontology;
import gate.util.GateRuntimeException;
import gate.util.InvalidOffsetException;
import gate.util.LuckyException;
import gate.util.Out;

import java.util.Locale;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.Set;

/**
 * GATE Processing Resource
 * 
 * @author Telematica Instituut, Christian Wartena
 */
public class Apolda extends AbstractLanguageAnalyser {

	public static final String LOOKUP_INSTANCE_FEATURE_NAME = "identifier";
	public static final String CONCEPT_FEATURE_NAME = "Mention";
	
	private static final String EXC_NOONT = "Please specify an ontology!";
	private static final String EXC_ABRUPT = "The execution of appolda has been abruptly interrupted!";
	private static final String EXC_NODOC = "No document to process!";
	private static final String EXC_NOTOKENS = "You need to run a Tokenizer first!";
	private static final String DOCPROP_NROFTOK = "Number of tokens";
	private static final String DOCPROP_NROFCONCEPTS = "Number of recognized concepts";
	private static final String STATUS_STARTDOC = "Doing ontology lookup in";
	private static final String STATUS_ENDDOC = "Ontology lookup complete!";
	private static final String STATUS_STARTREADONT = "Reading ontology";
	private static final String STATUS_ENDREADONT = "term representations read";
	
	private String inputASname, outputASname;

	private Locale language = null;
	
	private String lemmaFeature = null;

	private String sURL;

	private String prefRepresentation = null;
	
	private String altRepresentation = null;

	private Ontology ontology;

	/**
	 * The initial state of the FSM that backs this gazetteer
	 */
	protected FSMState initialState;

	/**
	 * A set containing all the states of the FSM backing the gazetteer
	 */
	protected Set fsmStates;

	public Apolda() {
	}

	public String getInputASName() {
		return inputASname;
	}

	public void setInputASName(String inputASname) {
		this.inputASname = inputASname;
	}

	public String getOutputASName() {
		return outputASname;
	}

	public void setOutputASName(String outputASname) {
		this.outputASname = outputASname;
	}
	
	public String getLemmaFeature() {
		return lemmaFeature;
	}

	public void setLemmaFeature(String lemmaFeature) {
		this.lemmaFeature = lemmaFeature;
	}

	public Resource init() throws ResourceInstantiationException {
		
			
		if (ontology == null) {
			throw new ResourceInstantiationException(EXC_NOONT);
		}
		
		fsmStates = new HashSet();
		initialState = new FSMState();
		fsmStates.add(initialState);
		readOntology();

		return this;
	}

	private int addAnnotation(AnnotationSet annotationSet, Set lookupSet, Region matchedRegion) {
		Term currentTerm;
		FeatureMap fm;
		int nrOfAnnotations = 0;
		Iterator lookupIter = lookupSet.iterator();

		while (lookupIter.hasNext()) {
			currentTerm = (Term) lookupIter.next();
			fm = Factory.newFeatureMap();
			
			fm.put(LOOKUP_CLASS_FEATURE_NAME, currentTerm.oClass);
			fm.put(LOOKUP_ONTOLOGY_FEATURE_NAME, currentTerm.ontology);
			if (!currentTerm.oIdentifier.equals("")) {
				fm.put(LOOKUP_INSTANCE_FEATURE_NAME, currentTerm.oIdentifier);
			}
			
			try {
				annotationSet.add(new Long(matchedRegion.getBegin()), new Long(
						matchedRegion.getEnd() + 1), CONCEPT_FEATURE_NAME, fm);
				nrOfAnnotations++;
			} catch (InvalidOffsetException ioe) {
				throw new LuckyException(ioe.toString());
			}
		}// while(lookupIter.hasNext())

		return nrOfAnnotations;
	}

	private int findConcepts(Content theContent,AnnotationSet outputAnnSet) 
	throws ExecutionException {	 
		int nrOfAnnotations = 0;
		int iterations = 0;
		char currentChar;
		FSMState currentState = initialState;
		FSMState nextState;
		FSMState lastMatchingState = null;
		Region lastMatchingRegion = null;
		Set termSet;
		
		while (theContent.more()) {
			currentChar = theContent.getNextChar(currentState);						
			currentChar = Character.toUpperCase(currentChar);
			
			nextState = currentState.next(currentChar);
			if (nextState == null) { // the matching stopped	
				// Try backtracking in order to find a longer match
				currentState = theContent.redo();
				if(currentState == null) {
					// reset the FSM
					currentState = initialState;
					theContent.forward(currentState);
			
					// if we had a successful match then act on it;
					if (lastMatchingState != null && lastMatchingRegion != null) {
						termSet = checkMatchingTerms(lastMatchingRegion,lastMatchingState.getTermSet());
						if(termSet != null) {
							nrOfAnnotations += addAnnotation(outputAnnSet, termSet,	lastMatchingRegion);
							lastMatchingState = null;
						}
					}
				}					
			} else {// go on with the matching
				currentState = nextState;
				// if we have a successful state then store it
				if (currentState.isFinal() && theContent.atEndOfWord()) {
					Region r = theContent.getCurrentRegion();
					if(lastMatchingRegion == null || !lastMatchingRegion.contains(r)) {
						lastMatchingState = currentState;
						lastMatchingRegion = r;
					}
				}
			} 
	
			if((iterations++)%256 == 0 && isInterrupted()) {
				throw new ExecutionInterruptedException(EXC_ABRUPT);
			}
				
		} //while (theContent.more()) 
		if (lastMatchingState != null) {
			termSet = checkMatchingTerms(lastMatchingRegion,lastMatchingState.getTermSet());
			if(termSet != null)
				nrOfAnnotations += addAnnotation(outputAnnSet,termSet, lastMatchingRegion);
		}
		
		return nrOfAnnotations;
	}
	
	private Set checkMatchingTerms(Region matchingRegion,Set termSet)
	throws ExecutionException {
		Set newTermSet = new HashSet(); 
		Iterator termSetIt = termSet.iterator();
		Term term;
		Long begin;
		Long end;
		String literalText = null;
		List representations;
		
		while(termSetIt.hasNext()) {
			term = (Term) termSetIt.next();
			if(term.type == Term.STANDARD) {
				newTermSet.add(term);
			} else {
				if(literalText == null) {
					try{
						begin = new Long(matchingRegion.getBegin());
						end = new Long(matchingRegion.getEnd() + 1);
						literalText = document.getContent().getContent(begin,end).toString();
					} catch(InvalidOffsetException e) {
						throw new ExecutionException(e);
					}
				}
				representations = term.getRepresentations();
				if(representations != null && representations.contains(literalText)) 
					newTermSet.add(term);
			}
		}
		return newTermSet;
	}
	
	public void execute() throws ExecutionException {
		int nrOfAnnotations = 0;

		try {
			// check the parameters
			if (document == null) {
				throw new GateRuntimeException(EXC_NODOC);
			}

			AnnotationSet inputAnnSet = inputASname == null
					|| inputASname.length() == 0 ? document.getAnnotations()
					: document.getAnnotations(inputASname);

			AnnotationSet outputAnnSet = outputASname == null
					|| outputASname.length() == 0 ? document.getAnnotations()
					: document.getAnnotations(outputASname);

			fireStatusChanged(STATUS_STARTDOC + " " + document.getName() + "...");

			try {
				document.getFeatures().put(
						DOCPROP_NROFTOK,
						new Integer(inputAnnSet.get(
								ANNIEConstants.TOKEN_ANNOTATION_TYPE).size())
								.toString());
			} catch (NullPointerException e) {
				throw new ExecutionException(EXC_NOTOKENS);
			}
			
			Content theContent = new Content(inputAnnSet,lemmaFeature);
			theContent.forward(initialState);
			
			nrOfAnnotations = findConcepts(theContent, outputAnnSet);
			document.getFeatures().put(DOCPROP_NROFCONCEPTS,
					new Integer(nrOfAnnotations).toString());
			
			fireProcessFinished();
			fireStatusChanged(STATUS_ENDDOC);

		} catch (Exception e) {
			throw new ExecutionException(e);
		}
	}

    private void addRepresentationToList(List representationsLit,
			List representationsStand, String representation) {
		int len;

		len = representation.length();
		if (len > 2 && representation.charAt(0) == '"'
				&& representation.charAt(len - 1) == '"') {
			representation = representation.substring(1, len - 1);
			representationsLit.add(representation);
		} else {
			representationsStand.add(representation);
		}
	}

	private void getRepresentations(OResource concept,
			List representationsLit, List representationsStand) {
		
		List propValL = null;
		Iterator propValIt;
		Literal representation;
		String strRepresentation;
		Locale lang;
		
		AnnotationProperty prefRepresentationAP = null;
		AnnotationProperty altRepresentationAP = null;	

		if (prefRepresentation == null && altRepresentation == null) {
			strRepresentation = concept.getName();
			if (strRepresentation != null) {
				strRepresentation = strRepresentation.trim();
				addRepresentationToList(representationsLit,
						representationsStand, strRepresentation);
			}
		}

		if (prefRepresentation != null) {
			prefRepresentationAP = (AnnotationProperty) ontology.getOResourceByName(prefRepresentation);
			if(prefRepresentationAP != null)
				propValL = concept.getAnnotationPropertyValues(prefRepresentationAP);
			if (propValL != null) {
				propValIt = propValL.iterator();
				while (propValIt.hasNext()) {
					representation = (Literal) propValIt.next();
					if(language != null){
						lang = representation.getLanguage();
					} else {
						lang = null;
					}
					if(lang == null || lang.getLanguage() == "" || lang.getLanguage() == language.getLanguage()){
						strRepresentation = representation.getValue();
						addRepresentationToList(representationsLit,
							representationsStand, strRepresentation);
					}
				}
			}
		}

		if (altRepresentation != null) {
			propValL = null;
			altRepresentationAP = (AnnotationProperty) ontology.getOResourceByName(altRepresentation);
			if(altRepresentationAP != null)	
				propValL = concept.getAnnotationPropertyValues(altRepresentationAP);
			if (propValL != null) {
				propValIt = propValL.iterator();
				while (propValIt.hasNext()) {
					representation = (Literal) propValIt.next();
					if(language != null){
						lang = representation.getLanguage();
					} else {
						lang = null;
					}
					if(lang == null || lang.getLanguage() == "" || lang.getLanguage() == language.getLanguage()){
						strRepresentation = representation.getValue();
						addRepresentationToList(representationsLit,
								representationsStand, strRepresentation);
					}
				}
			}
		}
	}
	
	private void readOntology() {
		fireStatusChanged(STATUS_STARTREADONT);
		// fireProgressChanged(x%);
		sURL = ontology.getURL().toString();
		int nrOfStrings = 0;

		Set concepts = ontology.getOClasses(false); 
		Iterator conceptIt = concepts.iterator();
		OClass concept;

		while (conceptIt.hasNext()) {
			concept = (OClass) conceptIt.next();
			Term termLit = new Term(concept.getName(), concept.getName(), sURL,Term.LITERAL);
			Term termGlob = new Term(concept.getName(), concept.getName(), sURL,Term.STANDARD);
			List representationsLit = new Vector();
			List representationsStandard = new Vector();
			getRepresentations(concept,representationsLit,representationsStandard);
			termLit.setRepresentations(representationsLit);
			nrOfStrings += addTermRepresentations(termLit,representationsLit);
			nrOfStrings += addTermRepresentations(termGlob,representationsStandard);
			nrOfStrings += readInstances(concept);
		}	
		Out.println(new Integer(nrOfStrings).toString() + " " + STATUS_ENDREADONT);
	}

	private int readInstances(OClass concept){
		Set instances = ontology.getOInstances(concept,OConstants.DIRECT_CLOSURE);
		Iterator instIt = instances.iterator();
		OInstance inst;
		int rc = 0;

		while (instIt.hasNext()) {
			inst = (OInstance) instIt.next();
			Term termLit = new Term(concept.getName(), inst.getName(), sURL,Term.LITERAL);
			Term termGlob = new Term(concept.getName(), inst.getName(), sURL,Term.STANDARD);
			List representationsLit = new Vector();
			List representationsStandard = new Vector();
			getRepresentations(inst,representationsLit,representationsStandard);
			termLit.setRepresentations(representationsLit);
			rc += addTermRepresentations(termLit,representationsLit);
			rc += addTermRepresentations(termGlob,representationsStandard);
		}
		return rc;
	}
	
	private int addTermRepresentations(Term term, List representations) {
		int rc = 0;
				
		if(representations != null) {
			Iterator representation = representations.iterator();
			while (representation.hasNext()) {
				addTermRepresentation((String) representation.next(), term);
				rc++;
			}
		}
	
		return rc;
	}
	
	private void addTermRepresentation(String representation, Term term) {
		char currentChar;
		FSMState currentState = initialState;
		FSMState nextState;
		boolean isSpace;

		for (int i = 0; i < representation.length(); i++) {
			currentChar = representation.charAt(i);
			isSpace = Character.isWhitespace(currentChar);
			if (isSpace) {
				if(i > 0 && Character.isWhitespace(representation.charAt(i-1)))
					continue;
				else
					currentChar = ' ';
			} else
				currentChar = Character.toUpperCase(currentChar);
			nextState = currentState.next(currentChar);
			if (nextState == null) {
				nextState = new FSMState();
				fsmStates.add(nextState);
				currentState.put(currentChar, nextState);
				if (isSpace)
					nextState.put(' ', nextState);
			}
			currentState = nextState;
		} // for(int i = 0; i< text.length(); i++)

		currentState.addTerm(term);
	} // addLookup


	public void setPrefRepresentation(String prefRepresentation) {
		this.prefRepresentation = prefRepresentation;
	}

	public String getPrefRepresentation() {
		return prefRepresentation;
	}
	
	public String getLanguage() {
	   return language.getLanguage();
	}

    public void setLanguage(String language) {
    	if(language != null && language != ""){
    		this.language = new Locale(language);
    	}
	  }

	public String getAltRepresentation() {
		return altRepresentation;
	}

	public void setAltRepresentation(String altRepresentation) {
		this.altRepresentation = altRepresentation;
	}

	public Ontology getOntology() {
		return ontology;
	}

	public void setOntology(Ontology ontology) {
		this.ontology = ontology;
	}
}